/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "GlobalSettings.hpp"
#include "BCPEngine.hpp"
#include "SharedBasic.hpp"
#include "SharedClause.hpp"
#include "SolverTrooper.hpp"
#include "DMRPTrooper.hpp"
#include "SolverLeader.hpp"
#include "ClauseInfos.hpp"
#include "Heap.hpp"
#include "Decision.hpp"
#include "RestartStrategies.hpp"
#include "AtomicTypes.hpp"
#include "VariableInfos.hpp"
#include "Watching.hpp"
#include "SolverFactory.hpp"
#include "BitVector.hpp"
#include "IFStream.hpp"
#include "LazyFraction.hpp"
#include "Logging.hpp"
#include "Random.hpp"
#include "SiblingInfo.hpp"
#include "OMPRun.hpp"

using std::cout; using std::endl;
const char*  globalsettings_cpp_version(){return "$Revision: 1.5 $";}

GlobalSettings::GlobalSettings(){

	//options for input
	Option<bool>::val_type val_true		= std::make_pair<const char*,bool>("yes",true);
	Option<bool>::val_type val_false	= std::make_pair<const char*,bool>("no",false);

	Option<bool>
	b_opt = Option<bool>(&print_console,    false,  			"-console=",  "If yes, some information is print to console output");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);

	Option<std::string>
	s_opt = Option<std::string>(&csv_file_name,   "",  		"-csv-out=",	"If specified, statistics are appended to this file");
	string_options.push_back(s_opt);

	s_opt = Option<std::string>(&dimacs_out,   "",  		"-dimacs-out=",	"If specified, the formula is printed to this file");
	string_options.push_back(s_opt);

	Option<uint>
	ui_opt = Option<uint>(&init_rand_seed,	 91648253,  	"-seed=",		"First random seed");
	uint_options.push_back(ui_opt);

	ui_opt = Option<uint>(&inc_rand_seed,	 2431234,  		"-seed-inc=",	"Increment to set initial seed for other threads");
	uint_options.push_back(ui_opt);

	/*ui_opt = Option<uint>(&max_lits,	 120000000,  		"-max-lits=",	"Forces Garbage Collection when this number of literals is reached");
	uint_options.push_back(ui_opt);

	ui_opt = Option<uint>(&bcp_comp_below,max32,  			"-bcp-comp=",	"Comp BCP is only applied below the given level");
	uint_options.push_back(ui_opt);*/

	ui_opt = Option<uint>(&timeout,		  	max32,			"-timeout=",	"The solver terminates after the given number of seconds");
	uint_options.push_back(ui_opt);

	uint t = min(8,omp_get_max_threads());
	ui_opt = Option<uint>(&nbr_threads, t, 					"-threads=",	"Number of threads to start");
	uint_options.push_back(ui_opt);


	/*ui_opt = Option<uint>(&ginit_rounds_print, 10000,		"-initprint=",	"Number of rounds in graph initialisation that have to pass until next stats are printed");
	uint_options.push_back(ui_opt);*/

}



bool GlobalSettings::read(int argc, char** argv){

	//check for correct usage
	solver_name = bname(argv[0]);
	if(argc < 2){
		fprintf(stderr,"Not enough Parameters! \n");
		return false;
	}

	int index(0);
	char* value;
	while(++index < argc){

		for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){
			if((value = bool_options[opt_idx].canHandle(argv[index]))){
				if(bool_options[opt_idx].findValue(value))
					goto OptionHandled; else return false;}
		}

		for(uint opt_idx = 0; opt_idx != uint_options.size();++opt_idx){
			if((value = uint_options[opt_idx].canHandle(argv[index]))){
				if(!uint_options[opt_idx].findValue(value)){
					uint_options[opt_idx].assign(atoi(value));}
				goto OptionHandled;
			}
		}

		for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){
			if((value = string_options[opt_idx].canHandle(argv[index]))){
				string_options[opt_idx].assign(value);
				goto OptionHandled;}
		}

		if(strncmp(argv[index], "-version", 8) == 0){
			printCVSVersionNumbers();
			exit(0);
		}

		if(*argv[index] == '-'){
			fprintf(stderr,"Unknown option: %s \n",argv[index]);
			return false;
		}else if(!cnf_file_name.empty()){
			std::cerr << "One of the two arguments is incorrect: \" " << cnf_file_name << "\" or \" " <<argv[index]<<"\""<<endl;
			return false;
		}else {
			cnf_file_name	= std::string(argv[index]);
		}

		OptionHandled: ;
	}

	uint32_t max_t	= min(8,omp_get_max_threads());
	if(nbr_threads>max_t){
		nbr_threads=max_t;
		std::cerr << "c OpenMP only supports " << nbr_threads << " threads on this machine";}

	return !cnf_file_name.empty();
};


void GlobalSettings::printUsage(){

	std::cout << "Program Usage:" << endl;
	std::cout << solver_name << " [Options] <dimacs-file> " << endl;
	std::cout << "Options are: " << endl;
	std::cout <<  std::setw(20) << "-version" << "\t\t" << "Displays all CVS Revision Numbers" << endl;
	for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){	bool_options[opt_idx].printDescription();}
	for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printDescription();}
	for(uint opt_idx = 0; opt_idx != uint_options.size();++opt_idx){	uint_options[opt_idx].printDescription();}
	std::cout << endl;

	std::cout << "Current Values are: " << endl;
	for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){	bool_options[opt_idx].printValue("%i");}
	for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printValue("%s");}
	for(uint opt_idx = 0; opt_idx != uint_options.size();++opt_idx){	uint_options[opt_idx].printValue("%u");}
}


#define HEADER(x,v) std::cout << std::setw(30) << x<<".hpp:   " << v##_hpp_version << endl;
#define SOURCE(x,v) std::cout << std::setw(30) << x<<".cpp:   " << v##_cpp_version() << endl;
#define HEADERSOURCE(x,v) HEADER(x,v) SOURCE(x,v)

void GlobalSettings::printCVSVersionNumbers(){

	std::cout << endl;
	std::cout << "SArTagnan" << endl;
#ifdef NDEBUG
	std::cout << "CVS Version Numbers (RELEASE Compilation)" << endl;
#else
	std::cout << "CVS Version Numbers (DEBUG Compilation)" << endl;
#endif
    SOURCE("SArTagnan",sapperlot);

    HEADER("AtomicTypes",sattypes);
    HEADER("BCPEngine",bcpengine);
    HEADER("ClauseInfos",clauseinfos);
    HEADER("Heap",heap);
    HEADER("Decision",decision);
    HEADER("RestartStrategies",restartstrategies);
    HEADER("Satistics",satistics);
    HEADER("SolverLeader",solverleader);
    HEADER("SolverTrooper",solvertrooper);
    HEADERSOURCE("DMRPTrooper",dmrptrooper);
    HEADER("VariableInfos",variableinfos);
    HEADER("Watching",watching);

    HEADER("LockFreeQueues",lockfreequeues);
    HEADER("OMPRun",omprun);
    HEADER("SharedBasic",sharedbasic);
    HEADER("SharedClause",sharedclause);
    HEADER("Watching",watching);
    HEADERSOURCE("SiblingInfo",siblinginfos);
    HEADERSOURCE("SolverFactory",solverfactory);
    HEADERSOURCE("GlobalSettings",globalsettings);

    HEADER("Assertion",assertion);
    HEADER("BitVector",bitvector);
    HEADER("IFStream",ifstream);
    HEADER("LazyFraction",lazyfraction);
    HEADER("Logging",logging);
    //HEADER("Queue",queue);
    //HEADER("UniqueContainer",uniquecontainer);
    HEADER("Random",random);
    //HEADER("VecOfVecs",vecofvecs);
    //HEADER("Info",info);

    //print_out("%s\n",utility_hpp_version);
    printSizes();
}


#define SIZEOF(x) std::cout << std::setw(30) << "Size of " << #x << ": " << sizeof(x) << endl;
void GlobalSettings::printSizes(){
	std::cout << endl << "Class Sizes " << endl;
	SIZEOF(void*)
	SIZEOF(ClsPointer)
	//SIZEOF(VarAct<BinTernVarInfo,Conflict2MaxLearnts>::act_t)
	SIZEOF(TrBool)
	SIZEOF(Variable)
	SIZEOF(Literal)
	SIZEOF(SharedClause)
	//SIZEOF(LevelValInfo)
	//SIZEOF(BTReason)
	SIZEOF(UserMask)
	SIZEOF(UserID)
	SIZEOF(UserBit)
	SIZEOF(UpdateChain<void>)



}





