/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef SETTINGS_HPP_
#define SETTINGS_HPP_


#define globalsettings_hpp_version "$Revision: 1.2 $"
const char*  globalsettings_cpp_version();
const char*  sapperlot_cpp_version();

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <vector>
#include <string>
#include <utility>
#include <sys/resource.h>
#include <fstream>
#include <algorithm>
#include <omp.h>

#include "Random.hpp"
#include "Logging.hpp"
#include "LazyFraction.hpp"
#include "Assertion.hpp"

const int EXIT_UNKNOWN 	= 0;
const int EXIT_SAT 		= 10;
const int EXIT_UNSAT 	= 20;

//#define NO_INTERRUPT

using std::endl; using std::cout;

//#define print_console(format, args...) ( fflush(stdout), fprintf(stderr, format, ## args), fflush(stderr) )


//user potions for command line
template<class T>
struct Option{
	typedef std::pair<const char*,T>		val_type;
private:
	T* 										variable;
	const char*								command_arg;
	int										command_len;
	const char*								description;
	std::vector<val_type >					possible_vals;

public:

	Option(T* var,T default_val,const char* arg,const char* descr)
		:variable(var),command_arg(arg),command_len(strlen(arg)),description(descr){
		*variable	= default_val;};

	void addPossibleValue(val_type pval){  possible_vals.push_back(pval);}

	char* canHandle(char* arg){
		return (strncmp(arg, command_arg, command_len) == 0) ? (arg + command_len) : NULL;}

	void assign(T val){*variable = val;}

	bool findValue(char* arg){
		if(possible_vals.empty()) return false;
		for(int i = possible_vals.size();i--;){
			if(strcmp(possible_vals[i].first,arg) == 0){
				assign(possible_vals[i].second);
				return true;
			}
		}
		std::cerr << "c Unknown value " << arg << " of option " << command_arg << endl;
		exit(EXIT_FAILURE);
		return false;
	}


	void printValue(std::string format){
		for(int i = possible_vals.size();i--;){
			if(*variable == possible_vals[i].second){
				std::cout << std::setw(20) << command_arg << "\t\t" << possible_vals[i].first << endl;
				return;
			}
		}
		std::cout << std::setw(20) << command_arg << "\t\t" <<*variable << endl;
	}


	void printDescription(){
		std::cout << std::setw(20) << command_arg << "\t\t" << description;
		int vals = possible_vals.size();
		if(vals){
			std::cout << " - possible values are: ";
			for(int i = 0; i != vals;++i) {std::cout << " | " << possible_vals[i].first;}
		}
		std::cout << std::endl;
	}
};


struct GlobalSettings{
	bool 			print_console;
	uint32_t		init_rand_seed;
	uint32_t		inc_rand_seed;
	//uint32_t		bcp_comp_below;
	uint32_t		timeout;
	//uint32_t		max_lits;
	//uint32_t		ginit_rounds_print;

	uint32_t		nbr_threads;
	//char*			threads_config_file;

	std::string		cnf_file_name;
	std::string		dimacs_out;
	std::string		csv_file_name;
	std::string		solver_name;

	std::vector<Option<uint> >  		uint_options;
	std::vector<Option<bool> >  		bool_options;
	std::vector<Option<std::string> > 	string_options;

	GlobalSettings();

	bool			read(int argc, char** argv);
	void			printUsage();
	void 			printCVSVersionNumbers();
	void 			printSizes();
};


const int KILO = 1 << 10;
const int MEGA = 1 << 20;


const uint16_t zero16 	=  0;
const uint32_t zero32 	=  0;
const uint64_t zero64 	=  0;
const uint16_t max16 	=  ~zero16;
const uint32_t max32 	=  ~zero32;
const uint64_t max64 	=  ~zero64;
const uint64_t one64 	=  1;

# if __WORDSIZE == 64
typedef uint64_t				ptr_int_t;
const ptr_int_t PTR_INT_BIT1	= 1;
const ptr_int_t PTR_INT_BIT2	= 2;
const ptr_int_t PTR_INT_MASK	= max64 - 3;
# else
typedef uint32_t		ptr_int_t;
const ptr_int_t PTR_INT_BIT1	= 1;
const ptr_int_t PTR_INT_BIT2	= 2;
const ptr_int_t PTR_INT_MASK	= max32 - 3;
#endif



inline void set_max_int(uint32_t& to_be_max, const uint32_t cand) { if(cand > to_be_max){to_be_max = cand;}}
inline void set_min_int(uint32_t& to_be_min, const uint32_t cand) { if(cand < to_be_min){to_be_min = cand;}}
inline uint32_t min(const uint32_t& a, const uint32_t& b){return a<b?a:b;}
inline uint32_t max(const uint32_t& a, const uint32_t& b){return a>b?a:b;}

inline uint32_t hash32(const int& x)					{ return 1 << (x & 31);}

inline uint64_t hash64(const int& x)					{ return one64 << (x & 63);}


template<class E, class Pred>
inline bool find(const std::vector<E>& vec,Pred p){

	int32_t i = vec.size();
	while(i--){if(p(vec[i])){ return true;}}

	return false;
}

template<class E>
inline void addIfNew(vector<E>& vec, const E& ele){

	for(uint32_t i=0;i!=vec.size();++i){if(vec[i]==ele) return;}
	vec.push_back(ele);
}



template<class E, class Pred>
inline void stable_remove_one(std::vector<E>& vec,Pred p){
	asrt(!vec.empty());

	int32_t i = vec.size();
	while(i--){if(p(vec[i])){ ++i; break;}}
	asrt( i > 0 );

	for(;i < (int32_t)vec.size();++i){vec[i-1]=vec[i];}
	vec.pop_back();
}


template<class E, class P>
inline void unstable_remove_one(std::vector<E>& vec, P p){
	asrt(!vec.empty());

	for(int32_t i(vec.size());i--;){
		if(p(vec[i])){
			vec[i] = vec.back(); vec.pop_back();
			return;
	}}
	asrt(false);
}


template<class E, class P>
inline void unstable_remove(E array[],uint32_t& sz, P p){
	for(int32_t i=0;i<sz;){
		if(p(array[i])){array[i] = array[--sz];}
		else{++i;}
	}
}

inline char* bname(char* str, char sep = '/'){
	for(char* iter = str; *iter != '\0' ; ){
		if(*iter++ == sep) str = iter;
	}
	return str;
}

/*
template<class T>
inline bool is_sorted(const T eles[],uint32_t strt, uint32_t sz){
	if(strt>=sz) return true;
	T last = eles[strt++];
	while(strt<sz){
		if(eles[strt]<last)	return false;
		last	= eles[strt++];
	}
	return true;
}

template<class T>
inline uint32_t bin_search(const T eles[],uint32_t st, uint32_t sz,const T& e){
	sanity(is_sorted(eles,st,sz));
	while(st<sz){
		const uint32_t p 	= (st+sz)>>1; asrt(p<sz);asrt(p>=st);

		if(eles[p]<e)		{st	=p+1;}
		else if(e<eles[p])	{sz	=p;}
		else {				return p;}
	}
	return max32;
}*/

/*
template<class T>
inline uint32_t bin_search(const vector<T>& eles,uint32_t st, uint32_t sz,const T& e){
	sanity(is_sorted(eles,sz));
	while(st<sz){
		const uint32_t p 	= (st+sz)>>1; asrt(p<sz);asrt(p>=st);

		if(eles[p]<e)		{st	=p+1;}
		else if(e<eles[p])	{sz	=p;}
		else {				return p;}
	}
	return max32;
}*/

template<class T,class I>
struct einfo: protected std::vector<T>{
	typedef std::vector<T> super;
	uint32_t size() const			{return super::size();}
	uint32_t capa() const			{return super::capacity();}
	bool inRange(const I& i)const 	{return i.idx() < size();}
	void reserve(const uint32_t& r){
		if(r>super::capacity())super::reserve(r);}
	void ensureSize(const uint32_t& s,const T& t=T()){
		asrt(s >= size());super::resize(s,t);}

	void ensureSizeExplicit(const uint32_t& s){
		asrt(s >= size());
		super::reserve(s); uint32_t idx;
		while((idx=size(),idx<s)){super::push_back(T(idx));}
	}

	const T& operator[](const I& i) const {
		asrt(inRange(i)); return super::operator [](i.idx());}
	T& operator[](const I& i) {
		asrt(inRange(i)); return super::operator [](i.idx());}

	const T& byIndex(const uint32_t& idx) const {asrt(idx < size());
		return super::operator[](idx);}
	T& byIndex(const uint32_t& idx) {asrt(idx < size());
		return super::operator[](idx);}

	I 	grow(const T& t){super::push_back(t);return I(size()-1);}
	void shrink(){asrt(!super::empty());super::pop_back();}
	//I	newUnused() const {return I(size());}
	void  setAll(const T& t){
		for(uint32_t i=size();i--;){super::operator[](i) = t;}}
};





#endif /*SETTINGS_HPP_*/
