/*
 * RestartStrategies.hpp
 *
 *  Created on: Sep 30, 2009
 *      Author: kottlers
 */

#ifndef RESTARTSTRATEGIES_HPP_
#define RESTARTSTRATEGIES_HPP_
#define restartstrategies_hpp_version      "$Revision: 1.5 $"

#include "GlobalSettings.hpp"
#include "HistoryBuffer.hpp"

enum RestartType{Luby,Geometric,Linear,NoRestarts,Glue};

class Restarts{
	const RestartType 	type;
	HistoryBuffer<uint8_t,uint64_t> lbd_history;


//luby related
	uint64_t 			modify;
	uint64_t 			last_pow2;
	uint64_t 			starts; //inited to 1
	uint64_t			confl_count;
	uint64_t			max;

//geometric: use "starts" as number of conflicts
	double 				confl_inc;

	volatile bool 		shall_run;

//linear -> use "starts" as number of conflicts and "modify" to increment

	Restarts(const RestartType rt,uint64_t f, uint64_t r, double i)
			:type(rt),modify(f),last_pow2(1),starts(r),confl_count(0)
			 ,max(0),confl_inc(i),shall_run(true){}

	Restarts(uint64_t buff_sz, uint64_t percent)
			:type(Glue),modify(percent),last_pow2(0),starts(buff_sz),confl_count(0)
			 ,max(0),confl_inc(0),shall_run(true){
			lbd_history.resize(buff_sz);}



	inline uint64_t getLuby();
	inline uint64_t getGeometric();
	inline uint64_t getLinear();

public:
	static Restarts createLuby(const uint32_t scale){
					return Restarts(Luby,scale,1,0);}
	static Restarts createGlucose(const uint32_t buff_sz, uint32_t prcnt){
					return Restarts(buff_sz,prcnt);}
	static Restarts createGeometric(const uint32_t start, double inc){
					return Restarts(Geometric,0,start,inc);}
	static Restarts createLinear(const uint64_t start, const uint64_t inc){
					return Restarts(Linear,inc,start,0);}
	static Restarts createStatic(const uint64_t freq){
					return createLinear(freq,0);}
	static Restarts createNoRestarts(){
					return createLinear(max64>>2,0);}

	inline void		nextStart();
	inline bool		incConflictBreak(const uint32_t& lbd);
	inline uint32_t getInitialNumber()const {return type == Luby ? modify : starts;}
	void			stopASAP(){shall_run = false;}
	bool			shallRun() const{return shall_run;}
};

void Restarts::nextStart(){
	switch(type){
		case Luby:{		max =  getLuby(); 		confl_count = 0;break;}
		case Geometric:{max =  getGeometric();	confl_count = 0;break;}
		case Glue:{		lbd_history.clear();	break;}
		default:{		max	=  getLinear();		confl_count = 0;}
	}

}

bool	Restarts::incConflictBreak(const uint32_t& lbd){
	++confl_count;
	if(type != Glue)	return confl_count > max ;

	lbd_history	.keep(lbd);
	max			+= lbd; //sum of all lbds

	return 		(lbd_history.enoughData() &&  ((lbd_history.getAvg()*modify) > ((100 *max) / confl_count)));
}

uint64_t Restarts::getLuby(){
	++starts;
	if((starts & -starts) == starts){ //power of 2
		last_pow2 = starts;
		return (starts >> 1)*modify;
	}

	//find value
	uint64_t start 	= starts;
	uint64_t l_pow2 = last_pow2;

	do{
		asrt(start > l_pow2 && start < 2* l_pow2);
		start -= (l_pow2 - 1);
		while(l_pow2 > start){l_pow2 >>= 1;}

	}while(start != l_pow2);

	return (start >> 1)*modify;
}

uint64_t Restarts::getLinear(){
	uint64_t r	= starts;
	starts 		+= modify;
	return 		r;
}

uint64_t Restarts::getGeometric(){
	uint64_t r	= starts;
	starts 		= (uint64_t)((double)starts * confl_inc);
	return 		r;
}


class ConflictsAndGC{
	//static const uint32_t 	maxconf	= ~zero32 - 300;
	uint32_t				nbr_orig_cls;

	uint32_t				inc_lrnts_limit_gc;
	uint32_t				inc_confl_limit_gc;

	uint32_t				deflt_lrnts_limit_gc;
	uint32_t				deflt_confl_limit_gc;

	uint32_t 				conflicts;

	uint32_t				max_confl_for_gc;
	uint32_t				max_learnts_for_gc;
	double 					shrink_on_gc;


public:
	ConflictsAndGC(uint32_t lrnts_allowed, uint32_t confl_allowed, uint32_t inc_lrnts_allow, uint32_t inc_confl_allow,double srnk_on_gc)
		:nbr_orig_cls			(0),
		inc_lrnts_limit_gc		(inc_lrnts_allow),
		inc_confl_limit_gc		(inc_confl_allow),
		deflt_lrnts_limit_gc	(lrnts_allowed),
		deflt_confl_limit_gc	(confl_allowed),
		conflicts				(0),
		max_confl_for_gc		(deflt_confl_limit_gc),
		max_learnts_for_gc		(deflt_lrnts_limit_gc),
		shrink_on_gc			(srnk_on_gc){}

	void setNumberClauses(uint32_t cls_count){
		nbr_orig_cls=cls_count;}

	uint32_t shallDoGC(const uint32_t& learnts_count, uint32_t lits){
		return (learnts_count < max_learnts_for_gc && conflicts < max_confl_for_gc && lits < 100000000)
				? max32 : (uint32_t)((double)learnts_count * shrink_on_gc);}

	void didGC(uint64_t curr_learnts){
		deflt_lrnts_limit_gc	+= inc_lrnts_limit_gc;
		deflt_confl_limit_gc	+= inc_confl_limit_gc;
		max_learnts_for_gc		=  curr_learnts+deflt_lrnts_limit_gc;
		max_confl_for_gc		=  conflicts+deflt_confl_limit_gc;}

	void incConflict(){	++conflicts;}
	uint64_t count() const 	{return conflicts;}

	/*void setToMax(){
		while(conflicts < maxconf){	conflicts = maxconf;}}

	static bool isMax(const uint64_t& conf) {return conf >= maxconf;}
	bool isMax() const {return isMax(conflicts);}*/
};


/*
class Conflicts2MaxLearnts{
	static const uint64_t 	maxconf	= ~zero64 - 100;
	static const double 	addend 	= 3;
	double 					cls_count100;
	uint64_t 				conflicts;
	int64_t   				log2;
	uint64_t				max_learnts;
	uint64_t				ignore_max_l_below_confl;
public:
	Conflicts2MaxLearnts():
		cls_count100(0),conflicts(0),log2(-1),max_learnts(0),ignore_max_l_below_confl(0){}

	void setNumberClauses(double cls_count){
		cls_count100 = cls_count/100.0;
		max_learnts	 = (uint64_t)(cls_count/3.0);
	}

	void ignoreMaxLearntsFor(uint32_t next_c){ignore_max_l_below_confl = conflicts + next_c;}

	void incConflict(){
		++conflicts;
		if((conflicts & -conflicts) == conflicts){ //power of 2
			++log2; asrt(((uint64_t)1 << (uint32_t)log2) == conflicts);
			double add_inc 	= (double)(log2/144.0);
			uint64_t cand	= (uint64_t)((double)log2*(addend+add_inc)*cls_count100);
			max_learnts		= cand > max_learnts ? cand : max_learnts;

			LOG_BASIC("Increment max_learnts to "<<max_learnts<<" at "<<conflicts<<" conflicts");
		}

	}

	uint64_t getMaxLearnts()const {return conflicts < ignore_max_l_below_confl ? max64 : max_learnts;}
	uint64_t count() const 	{return conflicts;}
	void setToMax() 			{
		while(conflicts < maxconf){
			conflicts = maxconf;
			//#pragma omp flush
		}
	}

	static bool isMax(const uint64_t& conf) {return conf == maxconf;}
	bool isMax() const {return isMax(conflicts);}
};

*/

#endif /* RESTARTSTRATEGIES_HPP_ */
